from logging import getLogger
from typing import Optional

import pandas as pd

LOG = getLogger("metloom.dataframe_utils")


def join_df(
    df: Optional[pd.DataFrame], new_df: Optional[pd.DataFrame], how="left", on=None
):
    """
    join two dataframes handling None
    Args:
        df: optional dataframe
        new_df: optional dataframe
        how: method for merging
        on: optional kwarg for DataFrame.join

    Returns:
        The joined dataframes. This method prefers values from the first if
        columns are overlapping and renames the overlapping values from
        the `new_df` to <column>_unused
    """
    if df is None:
        result_df = new_df
    elif new_df is None:
        result_df = df
    else:
        try:
            result_df = df.join(new_df, how=how, on=on, rsuffix="_unused")
        except Exception as e:
            LOG.error("failed joining dataframes.")
            raise e

    return result_df


def append_df(df: Optional[pd.DataFrame], new_df: Optional[pd.DataFrame]):
    """
    append 2 dfs handling Nones
    Args:
        df: optional dataframe
        new_df: optional dataframe
    Returns:
        dataframe or None
    """
    if df is None:
        result_df = new_df
    elif new_df is None:
        result_df = df
    else:
        result_df = df.append(new_df)
    return result_df



import numpy as np
import pytest
from unittest.mock import patch, MagicMock
from datetime import datetime, timezone, timedelta
from copy import deepcopy
import geopandas as gpd
import pandas as pd
import pytz
from os import path
from climata import snotel

from metloom.pointdata import (
    CDECPointData,
    PointDataCollection,
    PointData,
    SnotelPointData,
)
from metloom.variables import CdecStationVariables, SnotelVariables


def side_effect_error(*args):
    raise ValueError("Testing error")


class TestPointData(object):
    @pytest.fixture(scope="class")
    def data_dir(self):
        this_dir = path.dirname(__file__)
        return path.join(this_dir, "data")

    @pytest.fixture(scope="class")
    def shape_obj(self, data_dir):
        fp = path.join(data_dir, "testing.shp")
        return gpd.read_file(fp)

    @staticmethod
    def expected_response(dates, vals, var, unit, station, points):
        obj = []
        for dt, v in zip(dates, vals):
            obj.append(
                {
                    "datetime": pd.Timestamp(dt, tz="UTC"),
                    "measurementDate": pd.Timestamp(dt, tz="UTC"),
                    var: v,
                    f"{var}_units": unit,
                    "site": station.id,
                }
            )
        df = gpd.GeoDataFrame.from_dict(
            obj,
            geometry=[points] * len(dates),
        )
        # needed to reorder the columns for the pd testing compare
        df = df.filter(
            ["datetime", "geometry", "site", "measurementDate", var, f"{var}_units"]
        )
        df.set_index(keys=["datetime", "site"], inplace=True)
        return df


class TestCDECStation(TestPointData):
    @staticmethod
    def cdec_daily_response():
        return [
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 3,
                "sensorType": "SNOW WC",
                "date": "2021-5-16 00:00",
                "obsDate": "2021-5-16 00:00",
                "value": -0.11,
                "dataFlag": " ",
                "units": "INCHES",
            },
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 3,
                "sensorType": "SNOW WC",
                "date": "2021-5-17 00:00",
                "obsDate": "2021-5-17 00:00",
                "value": -0.10,
                "dataFlag": " ",
                "units": "INCHES",
            },
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 3,
                "sensorType": "SNOW WC",
                "date": "2021-5-18 00:00",
                "obsDate": "2021-5-18 00:00",
                "value": -0.10,
                "dataFlag": " ",
                "units": "INCHES",
            },
        ]

    @pytest.fixture(scope="function")
    def tny_station(self):
        return CDECPointData("TNY", "Tenaya Lake")

    @pytest.fixture(scope="class")
    def tny_daily_expected(self):
        points = gpd.points_from_xy([-119.0], [42.0], z=[1000.0])
        df = gpd.GeoDataFrame.from_dict(
            [
                {
                    "datetime": pd.Timestamp("2021-05-16 07:00:00+0000", tz="UTC"),
                    "measurementDate": pd.Timestamp(
                        "2021-05-16 07:00:00+0000", tz="UTC"
                    ),
                    "ACCUMULATED PRECIPITATION": -0.11,
                    "ACCUMULATED PRECIPITATION_units": "INCHES",
                    "site": "TNY",
                },
                {
                    "datetime": pd.Timestamp("2021-05-17 07:00:00+0000", tz="UTC"),
                    "measurementDate": pd.Timestamp(
                        "2021-05-17 07:00:00+0000", tz="UTC"
                    ),
                    "ACCUMULATED PRECIPITATION": -0.10,
                    "ACCUMULATED PRECIPITATION_units": "INCHES",
                    "site": "TNY",
                },
                {
                    "datetime": pd.Timestamp("2021-05-18 07:00:00+0000", tz="UTC"),
                    "measurementDate": pd.Timestamp(
                        "2021-05-18 07:00:00+0000", tz="UTC"
                    ),
                    "ACCUMULATED PRECIPITATION": -0.10,
                    "ACCUMULATED PRECIPITATION_units": "INCHES",
                    "site": "TNY",
                },
            ],
            geometry=[points[0]] * 3,
        )
        # needed to reorder the columns for the pd testing compare
        df = df.filter(
            [
                "datetime",
                "geometry",
                "site",
                "measurementDate",
                "ACCUMULATED PRECIPITATION",
                "ACCUMULATED PRECIPITATION_units",
            ]
        )
        df.set_index(keys=["datetime", "site"], inplace=True)
        return df

    @classmethod
    def tny_side_effect(cls, url, **kwargs):
        mock = MagicMock()
        if kwargs["params"].get("dur_code") == "D":
            mock.json.return_value = cls.cdec_daily_response()
        elif kwargs["params"].get("dur_code") == "H":
            raise NotImplementedError()
        elif "getStationInfo" in url:
            mock.json.return_value = {
                "STATION": [
                    {
                        "SENS_LONG_NAME": "SNOW, WATER CONTENT",
                        "ELEVATION": 1000.0,
                        "LATITUDE": 42.0,
                        "LONGITUDE": -119.0,
                    }
                ]
            }
        else:
            raise ValueError("unknown scenario")
        return mock

    @pytest.fixture(scope="class")
    def station_search_response(self):
        return [
            pd.DataFrame.from_records(
                [
                    (
                        "GIN",
                        "GIN FLAT",
                        "MERCED R",
                        "MARIPOSA",
                        -119.773,
                        37.767,
                        7050,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                    (
                        "DAN",
                        "DANA MEADOWS",
                        "TUOLUMNE R",
                        "TUOLUMNE",
                        -119.257,
                        37.897,
                        9800,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                    (
                        "TNY",
                        "TENAYA LAKE",
                        "MERCED R",
                        "MARIPOSA",
                        -119.448,
                        37.838,
                        8150,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                    (
                        "GFL",
                        "GIN FLAT (COURSE)",
                        "MERCED R",
                        "MARIPOSA",
                        -119.773,
                        37.765,
                        7000,
                        "Yosemite National Park",
                        np.nan,
                    ),
                    (
                        "TUM",
                        "TUOLUMNE MEADOWS",
                        "TUOLUMNE R",
                        "TUOLUMNE",
                        -119.350,
                        37.873,
                        8600,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                    (
                        "SLI",
                        "SLIDE CANYON",
                        "TUOLUMNE R",
                        "TUOLUMNE",
                        -119.43,
                        38.092,
                        9200,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                ],
                columns=[
                    "ID",
                    "Station Name",
                    "River Basin",
                    "County",
                    "Longitude",
                    "Latitude",
                    "ElevationFeet",
                    "Operator",
                    "Map",
                ],
            )
        ]

    def test_class_variables(self):
        assert CDECPointData("no", "no").tzinfo == pytz.timezone("US/Pacific")
        # Base implementation should fail
        with pytest.raises(AttributeError):
            PointData("foo", "bar").tzinfo

    def test_get_metadata(self, tny_station):
        with patch("metloom.pointdata.cdec.requests") as mock_requests:
            mock_requests.get.side_effect = self.tny_side_effect
            metadata = tny_station.metadata
            mock_get = mock_requests.get
            assert mock_get.call_count == 1
            mock_get.assert_called_with(
                "http://cdec.water.ca.gov/cdecstation2/CDecServlet/" "getStationInfo",
                params={"stationID": "TNY"},
            )
        expected = gpd.points_from_xy([-119.0], [42.0], z=[1000.0])[0]
        assert expected == metadata

    def test_get_daily_data(self, tny_station, tny_daily_expected):
        with patch("metloom.pointdata.cdec.requests") as mock_requests:
            mock_requests.get.side_effect = self.tny_side_effect
            response = tny_station.get_daily_data(
                datetime(2021, 5, 16),
                datetime(2021, 5, 18),
                [CdecStationVariables.PRECIPITATIONACCUM],
            )
            mock_get = mock_requests.get
            mock_get.assert_any_call(
                "http://cdec.water.ca.gov/dynamicapp/req/JSONDataServlet",
                params={
                    "Stations": "TNY",
                    "dur_code": "D",
                    "Start": "2021-05-16T00:00:00",
                    "End": "2021-05-18T00:00:00",
                    "SensorNums": "2",
                },
            )
            assert mock_get.call_count == 2
        pd.testing.assert_frame_equal(response, tny_daily_expected)

    def test_points_from_geometry(self, shape_obj, station_search_response):
        expected_url = (
            "https://cdec.water.ca.gov/dynamicapp/staSearch?"
            "sta=&sensor_chk=on&sensor=3"
            "&collect=NONE+SPECIFIED&dur="
            "&active_chk=on&active=Y"
            "&loc_chk=on&lon1=-119.8"
            "&lon2=-119.2&lat1=37.7"
            "&lat2=38.2"
            "&elev1=-5&elev2=99000&nearby=&basin=NONE+SPECIFIED"
            "&hydro=NONE+SPECIFIED&county=NONE+SPECIFIED"
            "&agency_num=160&display=sta"
        )
        with patch("metloom.pointdata.cdec.pd.read_html") as mock_table_read:
            mock_table_read.return_value = station_search_response
            result = CDECPointData.points_from_geometry(
                shape_obj, [CdecStationVariables.SWE]
            )
            mock_table_read.assert_called_with(expected_url)
            assert len(result) == 5
            assert [st.id for st in result] == ["GIN", "DAN", "TNY", "TUM", "SLI"]

    def test_points_from_geometry_fail(self, shape_obj):
        with patch("metloom.pointdata.cdec.pd") as mock_pd:
            mock_pd.read_html.side_effect = side_effect_error
            result = CDECPointData.points_from_geometry(
                shape_obj, [CdecStationVariables.SWE]
            )
            assert result.points == []

    def test_point_collection_to_dataframe(self, shape_obj, station_search_response):
        with patch("metloom.pointdata.cdec.pd.read_html") as mock_table_read:
            mock_table_read.return_value = station_search_response
            result = CDECPointData.points_from_geometry(
                shape_obj, [CdecStationVariables.SWE]
            )
            assert isinstance(result, PointDataCollection)
            points_df = result.to_dataframe()
            for idp, point in enumerate(result):
                point_row = points_df.iloc[idp]
                assert point.name == point_row["name"]
                assert point.id == point_row["id"]
                assert point.metadata == point_row["geometry"]


class MockSnotelIO(snotel.SnotelIO):
    """
    Mock the data structure that the climata.snotel classes return
    """

    def __init__(self, mock_obj, *args, **kwargs):
        """
        Args:
            mock_obj: a list of dictionaries that will be stored
        """
        self.mock_dict = mock_obj
        super(MockSnotelIO, self).__init__(*args, **kwargs)

    def load(self):
        self.data = self.mock_dict
        transformed = []
        # store the nested 'data' attribute as the same class
        # to allow attribute level access of variables
        if len(self.data) > 0 and self.data[0].get("data", False):
            for row in self.data:
                inner_data = MockSnotelIO(row["data"])
                new_row = deepcopy(row)
                new_row["data"] = inner_data
                transformed += [new_row]
            self.data = transformed


class TestSnotelPointData(TestPointData):
    @pytest.fixture(scope="function")
    def station(self):
        return SnotelPointData("538:CO:SNTL", "TestSite")

    @pytest.fixture(scope="class")
    def points(self):
        return gpd.points_from_xy([-107.67552], [37.9339], z=[9800.0])[0]

    @pytest.fixture(scope="class")
    def meta_response_dict(self):
        return {
            "actonId": "07M27S",
            "beginDate": "1979-10-01 00:00:00",
            "countyName": "Ouray",
            "elevation": 9800.0,
            "endDate": "2100-01-01 00:00:00",
            "fipsCountryCd": "US",
            "fipsCountyCd": "091",
            "fipsStateNumber": "08",
            "huc": "140200060201",
            "hud": "14020006",
            "latitude": 37.9339,
            "longitude": -107.67552,
            "name": "Idarado",
            "shefId": "IDRC2",
            "stationTriplet": "538:CO:SNTL",
        }

    @pytest.fixture(scope="class")
    def mock_meta_response(self, meta_response_dict):
        return MockSnotelIO([{"stationDataTimeZone": -8.0, **meta_response_dict}])

    @pytest.fixture(scope="class")
    def mock_coursemeta_response(self, meta_response_dict):
        return MockSnotelIO([meta_response_dict])

    @pytest.fixture(scope="function")
    def standard_snotel_return(self):
        return [{"elementCd": SnotelVariables.SWE.code, "storedunitcd": "in"}]

    def test_metadata(self, mock_meta_response):
        with patch("metloom.pointdata.snotel.snotel.StationMetaIO") as mock_snotel:
            mock_snotel.return_value = mock_meta_response
            obj = SnotelPointData("538:CO:SNTL", "eh")
            assert (
                obj.metadata
                == gpd.points_from_xy([-107.67552], [37.9339], z=[9800.0])[0]
            )
            assert obj.tzinfo == timezone(timedelta(hours=-8.0))

    @pytest.mark.parametrize(
        "mocked_class, date_name, dts, expected_dts, vals, d1, d2, fn_name",
        [
            (
                "snotel.StationHourlyDataIO",
                "dateTime",
                ["2020-03-20 00:00", "2020-03-20 01:00", "2020-03-20 02:00"],
                ["2020-03-20 08:00", "2020-03-20 09:00", "2020-03-20 10:00"],
                [13.19, 13.17, 13.14],
                datetime(2020, 3, 20, 0),
                datetime(2020, 3, 20, 2),
                "get_hourly_data",
            ),
            (
                "snotel.StationDailyDataIO",
                "date",
                ["2020-03-20", "2020-03-21", "2020-03-22"],
                ["2020-03-20 08:00", "2020-03-21 08:00", "2020-03-22 08:00"],
                [13.19, 13.17, 13.14],
                datetime(2020, 3, 20),
                datetime(2020, 3, 22),
                "get_daily_data",
            ),
            (
                "StationMonthlyDataIO",
                "date",
                ["2020-01-28", "2020-02-27"],
                ["2020-01-28 00:00", "2020-02-27 00:00"],
                [13.19, 13.17],
                datetime(2020, 1, 20),
                datetime(2020, 3, 15),
                "get_snow_course_data",
            ),
        ],
    )
    def test_get_data_methods(
        self,
        mocked_class,
        date_name,
        dts,
        expected_dts,
        vals,
        d1,
        d2,
        fn_name,
        station,
        standard_snotel_return,
        mock_meta_response,
        mock_coursemeta_response,
        points,
    ):
        with patch(f"metloom.pointdata.snotel.{mocked_class}") as mock_snotel, patch(
            "metloom.pointdata.snotel.snotel.StationMetaIO"
        ) as mock_meta:
            if "snow_course" in fn_name:
                mock_meta.return_value = mock_coursemeta_response
            else:
                mock_meta.return_value = mock_meta_response
            standard_snotel_return[0]["data"] = [
                {date_name: dt, "flag": "V", "value": v} for dt, v, in zip(dts, vals)
            ]
            mock_snotel.return_value = MockSnotelIO(standard_snotel_return)

            vrs = [SnotelVariables.SWE]
            fn = getattr(station, fn_name)
            result = fn(d1, d2, vrs)
            expected = self.expected_response(
                expected_dts, vals, SnotelVariables.SWE.name, "in", station, points
            )
            pd.testing.assert_frame_equal(result, expected)

    def test_points_from_geometry(self, shape_obj):
        with patch("metloom.pointdata.snotel.LoomStationIO") as mock_search:
            mock_search.return_value = MockSnotelIO(
                [
                    {
                        "actonId": None,
                        "beginDate": "1930-02-01 00:00:00",
                        "countyName": "Tuolumne",
                        "elevation": 6500.0,
                        "endDate": "2100-01-01 00:00:00",
                        "fipsCountryCd": "US",
                        "fipsCountyCd": "109",
                        "fipsStateNumber": "06",
                        "huc": "180400090302",
                        "latitude": 37.995,
                        "longitude": -119.78,
                        "name": "Fake1",
                        "shefId": None,
                        "stationTriplet": "FFF:CA:SNOW",
                    },
                    {
                        "actonId": None,
                        "beginDate": "1948-02-01 00:00:00",
                        "countyName": "Tuolumne",
                        "elevation": 9300.0,
                        "endDate": "2100-01-01 00:00:00",
                        "fipsCountryCd": "US",
                        "fipsCountyCd": "109",
                        "fipsStateNumber": "06",
                        "huc": "180400090402",
                        "hud": "18040009",
                        "latitude": 38.18333,
                        "longitude": -119.61667,
                        "name": "Fake2",
                        "shefId": None,
                        "stationTriplet": "BBB:CA:SNOW",
                    },
                ]
            )
            result = SnotelPointData.points_from_geometry(
                shape_obj, [SnotelVariables.SWE], snow_courses=True
            )
            ids = [point.id for point in result]
            names = [point.name for point in result]
            assert ids == ["FFF:CA:SNOW", "BBB:CA:SNOW"]
            assert names == ["Fake1", "Fake2"]

    def test_points_from_geometry_fail(self, shape_obj):
        with patch("metloom.pointdata.snotel.LoomStationIO") as mock_search:
            mock_search.return_value = MockSnotelIO([])
            result = SnotelPointData.points_from_geometry(
                shape_obj, [SnotelVariables.SWE], snow_courses=True
            )
            assert len(result) == 0



from .base import PointData, PointDataCollection
from .cdec import CDECPointData
from .snotel import SnotelPointData

__all__ = ["PointData", "PointDataCollection", "CDECPointData", "SnotelPointData"]



from datetime import datetime
from typing import List

import geopandas as gpd
import numpy as np
import pandas as pd
import pytz
import requests
import logging

from .base import PointData
from ..variables import CdecStationVariables, SensorDescription
from ..dataframe_utils import join_df

LOG = logging.getLogger("metloom.pointdata.cdec")


class CDECPointData(PointData):
    """
    Implement PointData methods for CDEC data source
    API documentation here https://cdec.water.ca.gov/dynamicapp/
    """

    ALLOWED_VARIABLES = CdecStationVariables
    CDEC_URL = "http://cdec.water.ca.gov/dynamicapp/req/JSONDataServlet"
    META_URL = "http://cdec.water.ca.gov/cdecstation2/CDecServlet/" "getStationInfo"

    def __init__(self, station_id, name, metadata=None):
        """
        See docstring for PointData.__init__
        """
        super(CDECPointData, self).__init__(station_id, name, metadata=metadata)
        self._raw_metadata = None
        self._tzinfo = pytz.timezone("US/Pacific")

    def _get_all_metadata(self):
        """
        Get all the raw metadata for a station. This is a list of sensor
        descriptions for the station
        Returns:
            A list of dictionaries describing the sensors at a station
        """
        if self._raw_metadata is None:
            resp = requests.get(self.META_URL, params={"stationID": self.id})
            resp.raise_for_status()
            self._raw_metadata = resp.json()["STATION"]
        return self._raw_metadata

    def is_only_snow_course(self):
        """
        Determine if a station only has snow course measurements
        """
        data = self._get_all_metadata()
        manual_check = [d["DUR_CODE"] == "M" for d in data if d["SENS_GRP"] == "snow"]
        result = False
        if len(manual_check) > 0 and all(manual_check):
            result = True
        if result and not self.is_only_monthly():
            # This would happen if all snow sensors have code "M"
            # but there are other hourly or daily sensors
            raise Exception(
                f"We have not accounted for this scenario. Please talk to "
                f"a Micah about how {self.id} violates their assumptions."
            )
        return result

    def is_partly_snow_course(self):
        """
        Determine if any of the snow sensors at a station are on a monthly
        interval
        Assumption: Monthly snow sensor measurements are snow courses
        """
        data = self._get_all_metadata()
        return any([d["DUR_CODE"] == "M" for d in data if d["SENS_GRP"] == "snow"])

    def is_only_monthly(self):
        """
        determine if all sensors for a station are on a monthly interval
        """
        data = self._get_all_metadata()
        manual_check = [d["DUR_CODE"] == "M" for d in data]
        if len(manual_check) > 0 and all(manual_check):
            return True
        return False

    def _get_metadata(self):
        """
        See docstring for PointData._get_metadata
        """
        data = self._get_all_metadata()
        # TODO: Should this be sensor specific?
        metadata_by_name = {d["SENS_LONG_NAME"]: d for d in data}
        # default to the first sensor
        chosen_sensor_data = data[0]
        # try to replace it with a desired sensor
        # TODO: Change this
        for choice in [
            "SNOW, WATER CONTENT",
            "SNOW DEPTH",
            "PRECIPITATION, ACCUMULATED",
        ]:
            if metadata_by_name.get(choice) is not None:
                chosen_sensor_data = metadata_by_name[choice]
                break

        return gpd.points_from_xy(
            [chosen_sensor_data["LONGITUDE"]],
            [chosen_sensor_data["LATITUDE"]],
            z=[chosen_sensor_data["ELEVATION"]],
        )[0]

    def _data_request(self, params):
        """
        Make get request to CDEC and return JSON
        Args:
            params: dictionary of request parameters
        Returns:
            dictionary of response values
        """
        resp = requests.get(self.CDEC_URL, params=params)
        resp.raise_for_status()
        return resp.json()

    def _sensor_response_to_df(self, response_data, sensor, final_columns):
        """
        Convert the response data from the API to a GeoDataFrame
        Format and map columns in the dataframe
        Args:
            response_data: JSON list response from CDEC API
            sensor: SensorDescription obj
            final_columns: List of columns used for filtering
        Returns:
            GeoDataFrame
        """
        sensor_df = gpd.GeoDataFrame.from_dict(
            response_data,
            geometry=[self.metadata] * len(response_data),
        )
        sensor_df.replace(-9999.0, np.nan, inplace=True)
        # this mapping is important. Sometimes obsDate is null
        sensor_df.rename(
            columns={
                "date": "datetime",
                "obsDate": "measurementDate",
                "value": sensor.name,
                "units": f"{sensor.name}_units",
                "stationId": "site",
            },
            inplace=True,
        )
        final_columns += [sensor.name, f"{sensor.name}_units"]
        sensor_df["datetime"] = pd.to_datetime(sensor_df["datetime"])
        sensor_df["measurementDate"] = pd.to_datetime(sensor_df["measurementDate"])
        sensor_df["datetime"] = sensor_df["datetime"].apply(self._handle_df_tz)
        sensor_df["measurementDate"] = sensor_df["measurementDate"].apply(
            self._handle_df_tz
        )
        # set index so joinng works
        sensor_df.set_index("datetime", inplace=True)
        sensor_df = sensor_df.filter(final_columns)
        return sensor_df

    def _get_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
        duration: str,
    ):
        """
        Args:
            start_date: datetime object for start of data collection period
            end_date: datetime object for end of data collection period
            variables: List of metloom.variables.SensorDescription object
                from self.ALLOWED_VARIABLES
            duration: CDEC duration code ['M', 'H', 'D']
        Returns:
            GeoDataFrame of data, indexed on datetime, site
        """
        params = {
            "Stations": self.id,
            "dur_code": duration,
            "Start": start_date.isoformat(),
            "End": end_date.isoformat(),
        }
        df = None
        final_columns = ["geometry", "site", "measurementDate"]
        for sensor in variables:
            params["SensorNums"] = sensor.code
            response_data = self._data_request(params)
            if response_data:
                sensor_df = self._sensor_response_to_df(
                    response_data, sensor, final_columns
                )
                df = join_df(df, sensor_df)

        if df is not None and len(df.index) > 0:
            df.reset_index(inplace=True)
            df.set_index(keys=["datetime", "site"], inplace=True)
            df.index.set_names(["datetime", "site"], inplace=True)
        self.validate_sensor_df(df)
        return df

    def get_daily_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_daily_data
        Example query:
        https://cdec.water.ca.gov/dynamicapp/req/JSONDataServlet?
        Stations=TNY&SensorNums=3&dur_code=D&Start=2021-05-16&End=2021-05-16
        """
        return self._get_data(start_date, end_date, variables, "D")

    def get_hourly_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_hourly_data
        """
        return self._get_data(start_date, end_date, variables, "H")

    def get_snow_course_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_snow_course_data
        """
        if not self.is_partly_snow_course():
            raise ValueError(f"{self.id} is not a snow course")
        return self._get_data(start_date, end_date, variables, "M")

    @staticmethod
    def _station_sensor_search(
        bounds, sensor: SensorDescription, dur=None, collect=None
    ):
        """
        Station search form https://cdec.water.ca.gov/dynamicapp/staSearch?
        Search for stations using the CDEC station search utility
        Args:
            bounds: dictionary of Longitude and Latitidue bounds with keys
                minx, maxx, miny, maxy
            sensor: SensorDescription object
            dur: optional CDEC duration code ['M', 'H', 'D']
            collect: optional CDEC collection type string i.e. 'MANUAL+ENTRY'
        Returns:
            Pandas Dataframe of table result or None if no table found

        """
        # TODO: do we want this buffer?
        buffer = 0.00
        dur_str = f"&dur_chk=on&dur={dur}" if dur else "&dur="
        collect_str = (
            f"&collect_chk=on&collect={collect}"
            if collect
            else "&collect=NONE+SPECIFIED"
        )
        url = (
            f"https://cdec.water.ca.gov/dynamicapp/staSearch?sta="
            f"&sensor_chk=on&sensor={sensor.code}"
            f"{collect_str}"
            f"{dur_str}"
            f"&active_chk=on&active=Y"
            f"&loc_chk=on"
            f"&lon1={bounds['minx']-buffer}&lon2={bounds['maxx']+buffer}"
            f"&lat1={bounds['miny']-buffer}&lat2={bounds['maxy']+buffer}"
            f"&elev1=-5&elev2=99000&nearby=&basin=NONE+SPECIFIED"
            f"&hydro=NONE+SPECIFIED&county=NONE+SPECIFIED&agency_num=160"
            f"&display=sta"
        )
        try:
            return pd.read_html(url)[0]
        except ValueError:
            LOG.error(f"No tables for {url}")
            return None

    @classmethod
    def points_from_geometry(
        cls,
        geometry: gpd.GeoDataFrame,
        variables: List[SensorDescription],
        snow_courses=False,
    ):
        """
        See docstring for PointData.points_from_geometry
        """
        # Assume station search result is in 4326
        projected_geom = geometry.to_crs(4326)
        bounds = projected_geom.bounds.iloc[0]
        search_df = None
        station_search_kwargs = {}
        # Filter to manual, monthly measurements if looking for snow courses
        if snow_courses:
            station_search_kwargs["dur"] = "M"
            station_search_kwargs["collect"] = "MANUAL+ENTRY"
        for variable in variables:
            result_df = cls._station_sensor_search(
                bounds, variable, **station_search_kwargs
            )
            if result_df is not None:
                result_df["index_id"] = result_df["ID"]
                result_df.set_index("index_id")
                search_df = join_df(search_df, result_df, how="outer")
        # return empty collection if we didn't find any points
        if search_df is None:
            return cls.ITERATOR_CLASS([])
        gdf = gpd.GeoDataFrame(
            search_df,
            geometry=gpd.points_from_xy(
                search_df["Longitude"],
                search_df["Latitude"],
                z=search_df["ElevationFeet"],
            ),
        )
        # filter to points within shapefile
        # TODO: do we want to make this optional?
        filtered_gdf = gdf[gdf.within(projected_geom.iloc[0]["geometry"])]

        points = [
            cls(row[0], row[1], metadata=row[2])
            for row in zip(
                filtered_gdf["ID"],
                filtered_gdf["Station Name"],
                filtered_gdf["geometry"],
            )
        ]
        # filter to snow courses or not snowcourses depending on desired result
        if snow_courses:
            return cls.ITERATOR_CLASS([p for p in points if p.is_partly_snow_course()])
        else:
            return cls.ITERATOR_CLASS(
                [p for p in points if not p.is_only_snow_course()]
            )



from datetime import datetime, timezone, timedelta
from typing import List

import logging
import geopandas as gpd
import pandas as pd
from climata import snotel
from climata.base import FilterOpt, DateOpt, as_list
from wq.io.exceptions import NoData

from .base import PointData
from ..variables import SnotelVariables, VariableBase, SensorDescription
from ..dataframe_utils import join_df


LOG = logging.getLogger("metloom.pointdata.snotel")


class LoomStationIO(snotel.StationIO):
    """
    Extend climata.snotel.StationIO to allow filtering based on network
    and longitude
    """

    max_longitude = FilterOpt(url_param="maxLongitude")
    min_longitude = FilterOpt(url_param="minLongitude")
    networks = FilterOpt(url_param="networkCds")


class MonthlyDataIO(snotel.SnotelIO):
    """
    Extend climata.snotel.SnotelIO for sampling SEMIMONTHLY data.
    This allows sampling of snowcourses
    """

    data_function = "getData"

    # Applicable WebserviceLoader default options
    station = FilterOpt(required=True, url_param="stationTriplets")
    parameter = FilterOpt(required=True, url_param="elementCd")
    start_date = DateOpt(required=True, url_param="beginDate")
    end_date = DateOpt(required=True, url_param="endDate")

    default_params = {
        "ordinal": 1,
        "duration": "SEMIMONTHLY",
        "getFlags": "true",
        "alwaysReturnDailyFeb29": "false",
    }

    def parse(self):
        data = self.data[0]
        if not data or "values" not in data:
            raise NoData
        dates = as_list(data["collectionDates"])
        vals = as_list(data["values"])
        flags = as_list(data["flags"])

        self.data = [
            {
                "date": datetime.strptime(date_val, "%Y-%m-%d").date(),
                "value": val,
                "flag": flag,
            }
            for date_val, val, flag in zip(dates, vals, flags)
            if date_val is not None
        ]


class StationMonthlyDataIO(snotel.StationDataIO):
    """
    Extension of climata.snotel.StationDataIO to allow snowcourse sampling
    """

    inner_io_class = MonthlyDataIO
    duration = "SEMIMONTHLY"


class SnotelPointData(PointData):
    """
    Implement PointData methods for SNOTEL data source
    API documentation here:
    https://www.nrcs.usda.gov/wps/portal/wcc/home/dataAccessHelp/webService/webServiceReference/
    Website has variable and network codes
    Possible testing resource:
    https://wcc.sc.egov.usda.gov/awdbWebService/webservice/testwebservice.jsf?webserviceName=/awdbWebService
    """

    ALLOWED_VARIABLES = SnotelVariables

    def __init__(self, station_id, name, metadata=None):
        """
        See docstring for PointData.__init__
        """
        super(SnotelPointData, self).__init__(station_id, name, metadata=metadata)
        self._raw_metadata = None
        self._tzinfo = None

    def _snotel_response_to_df(
        self,
        response: snotel.StationDataIO,
        variables: List[VariableBase],
        date_key="date",
    ):
        """
        Convert the response from climata.snotel classes into
        Args:
            response: climata.snotel data response
            variables: list of variables to extract
            date_key: which key the response uses to store date(time)s
        """
        df = None
        # TODO: possible DRY oportunity here too
        final_columns = ["geometry", "site", "measurementDate"]
        for param in response:
            for variable in variables:
                if param.elementcd == variable.code:
                    data = param.data
                    transformed = [
                        {
                            "datetime": getattr(v, date_key),
                            "site": self.id,
                            "measurementDate": getattr(v, date_key),
                            variable.name: v.value,
                            f"{variable.name}_units": param.storedunitcd,
                        }
                        for v in data
                    ]
                    final_columns += [variable.name, f"{variable.name}_units"]
                    sensor_df = gpd.GeoDataFrame.from_dict(
                        transformed, geometry=[self.metadata] * len(transformed)
                    )
                    # TODO: possibly an opportunity for DRY here (see CDEC)
                    sensor_df["datetime"] = pd.to_datetime(sensor_df["datetime"])
                    sensor_df["measurementDate"] = pd.to_datetime(
                        sensor_df["measurementDate"]
                    )
                    sensor_df["datetime"] = sensor_df["datetime"].apply(
                        self._handle_df_tz
                    )
                    sensor_df["measurementDate"] = sensor_df["measurementDate"].apply(
                        self._handle_df_tz
                    )
                    # set index so joining works
                    sensor_df.set_index("datetime", inplace=True)
                    sensor_df = sensor_df.filter(final_columns)
                    df = join_df(df, sensor_df)

        if df is not None and len(df.index) > 0:
            df.reset_index(inplace=True)
            df.set_index(keys=["datetime", "site"], inplace=True)
            df.index.set_names(["datetime", "site"], inplace=True)
        self.validate_sensor_df(df)
        return df

    def get_daily_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_daily_data
        """
        params = snotel.StationDailyDataIO(
            station=self.id,
            start_date=start_date.date().isoformat(),
            end_date=end_date.date().isoformat(),
        )
        return self._snotel_response_to_df(params, variables)

    def get_hourly_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_hourly_data
        """
        params = snotel.StationHourlyDataIO(
            station=self.id,
            start_date=start_date.date().isoformat(),
            end_date=end_date.date().isoformat(),
        )
        return self._snotel_response_to_df(params, variables, date_key="datetime")

    def get_snow_course_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_snow_course_data
        """
        params = StationMonthlyDataIO(
            station=self.id,
            start_date=start_date.date().isoformat(),
            end_date=end_date.date().isoformat(),
        )
        return self._snotel_response_to_df(params, variables, date_key="date")

    def _get_all_metadata(self):
        """
        Set _raw_metadata once using climata.snotel
        """
        if self._raw_metadata is None:
            self._raw_metadata = snotel.StationMetaIO(station=self.id)
        return self._raw_metadata

    def _get_metadata(self):
        """
        See docstring for PointData._get_metadata
        """
        all_metadata = self._get_all_metadata()
        data = all_metadata[0]
        return gpd.points_from_xy(
            [data.longitude], [data.latitude], z=[data.elevation]
        )[0]

    def _get_tzinfo(self):
        """
        Return timezone info that pandas can use from the raw_metadata
        """
        all_metadata = self._get_all_metadata()
        metadata = all_metadata[0]
        # Snow courses might not have a timezone attached
        if hasattr(metadata, "stationdatatimezone"):
            tz_hours = metadata.stationdatatimezone
        else:
            LOG.error(f"Could not find timezone info for {self.id} ({self.name})")
            tz_hours = 0
        return timezone(timedelta(hours=tz_hours))

    @property
    def tzinfo(self):
        if self._tzinfo is None:
            self._tzinfo = self._get_tzinfo()
        return self._tzinfo

    @classmethod
    def points_from_geometry(
        cls,
        geometry: gpd.GeoDataFrame,
        variables: List[SensorDescription],
        snow_courses=False,
    ):
        """
        See docstring for PointData.points_from_geometry
        """
        projected_geom = geometry.to_crs(4326)
        bounds = projected_geom.bounds.iloc[0]
        # TODO: network may need to change to get streamflow
        network = "SNOW" if snow_courses else "SNTL"
        df = None
        for variable in variables:
            # this search is default AND on all parameters
            # so search for each variable seperately
            response = LoomStationIO(
                max_latitude=bounds["maxy"],
                min_latitude=bounds["miny"],
                max_longitude=bounds["maxx"],
                min_longitude=bounds["minx"],
                networks=network,  # only snotel, not
                parameter=variable.code,
                # ordinals=  # TODO: what are ordinals?
            )
            if len(response.data) > 0:
                result_df = pd.DataFrame.from_records(response.data)
                result_df["index_id"] = result_df["stationTriplet"]
                df = join_df(df, result_df, how="outer")
        # TODO: possible DRYing out here

        if df is None:
            return cls.ITERATOR_CLASS([])

        gdf = gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(
                df["longitude"], df["latitude"], z=df["elevation"]
            ),
        )
        filtered_gdf = gdf[gdf.within(projected_geom.iloc[0]["geometry"])]
        points = [
            cls(row[0], row[1], metadata=row[2])
            for row in zip(
                filtered_gdf["stationTriplet"],
                filtered_gdf["name"],
                filtered_gdf["geometry"],
            )
        ]
        return cls.ITERATOR_CLASS(points)



from dataclasses import dataclass


@dataclass
class SensorDescription:
    """
    data class for describing a snow sensor
    """

    code: str = "-1"  # code used within the applicable API
    name: str = "basename"  # desired name for the sensor
    description: str = None  # description of the sensor
    accumulated: bool = False  # whether or not the data is accumulated


class VariableBase:
    """
    Base class to store all variables for a specific datasource. Each
    datasource should implement the class. The goal is that the variables
    are synonymous across implementations.(i.e. PRECIPITATION should have the
    same meaning in each implementation).
    Variables in this base class should ideally be implemented by all classes
    and cannot be directly used from the base class.
    """

    PRECIPITATION = SensorDescription()
    SWE = SensorDescription()
    SNOWDEPTH = SensorDescription()

    @staticmethod
    def _validate_sensor(sensor: SensorDescription):
        """
        Validate that a sensor is not using the default values since they
        are meaningless
        """
        default = SensorDescription()
        if sensor.name == default.name and sensor.code == default.code:
            raise ValueError(f"{sensor.name} is the default implementation")

    @classmethod
    def from_code(cls, code):
        """
        Get the correct sensor description from the code
        """
        for k, v in cls.__dict__.items():
            if isinstance(v, SensorDescription) and v.code == str(code):
                cls._validate_sensor(v)
                return v
        raise ValueError(f"Could not find sensor for code {code}")


class CdecStationVariables(VariableBase):
    """
    Available sensors from CDEC.
    Exhaustive list:
    http://cdec4gov.water.ca.gov/reportapp/javareports?name=SensList
    """

    PRECIPITATIONACCUM = SensorDescription(
        "2", "ACCUMULATED PRECIPITATION", "PRECIPITATION, ACCUMULATED", True
    )
    PRECIPITATION = SensorDescription(
        "45", "PRECIPITATION", "PRECIPITATION, INCREMENTAL", False
    )
    SNOWDEPTH = SensorDescription("18", "SNOWDEPTH", "SNOW DEPTH")
    SWE = SensorDescription("3", "SWE", "SNOW, WATER CONTENT", True)
    TEMP = SensorDescription("4", "AIR TEMP", "TEMPERATURE, AIR")
    TEMPAVG = SensorDescription("30", "AVG AIR TEMP", "TEMPERATURE, AIR AVERAGE")
    TEMPMIN = SensorDescription("32", "MIN AIR TEMP", "TEMPERATURE, AIR MINIMUM")
    TEMPMAX = SensorDescription("31", "MAX AIR TEMP", "TEMPERATURE, AIR MAXIMUM")


class SnotelVariables(VariableBase):
    """
    Available sensors from SNOTEL
    """

    SNOWDEPTH = SensorDescription("SNWD", "SNOWDEPTH")
    SWE = SensorDescription("WTEQ", "SWE")
    TEMP = SensorDescription("TOBS", "AIR TEMP")
    TEMPAVG = SensorDescription("TAVG", "AVG AIR TEMP",
                                "AIR TEMPERATURE AVERAGE")
    TEMPMIN = SensorDescription("TMIN", "MIN AIR TEMP",
                                "AIR TEMPERATURE MINIMUM")
    TEMPMAX = SensorDescription("TMAX", "MAX AIR TEMP",
                                "AIR TEMPERATURE MAXIMUM")
    PRECIPITATION = SensorDescription("PRCPSA", "PRECIPITATON",
                                      "PRECIPITATION INCREMENT SNOW-ADJUSTED")
    PRECIPITATIONACCUM = SensorDescription("PREC", "ACCUMULATED PRECIPITATION",
                                           "PRECIPITATION ACCUMULATION")



import pandas as pd
import pytest

from metloom.dataframe_utils import join_df, append_df

df1 = pd.DataFrame.from_records([{"foo": 12.0}, {"foo": 11.0}])
df2 = pd.DataFrame.from_records([{"bar": 1.0}, {"bar": 1.0}])
df3 = pd.DataFrame.from_records([{"foo": 12.0, "bar": 1.0}, {"foo": 11.0, "bar": 1.0}])
df4 = pd.DataFrame.from_records(
    [
        {"foo": 12.0},
        {"foo": 11.0},
        {"foo": 12.0},
        {"foo": 11.0},
    ],
    index=[0, 1, 0, 1],
)


@pytest.mark.parametrize(
    "set1, set2, expected", [(df1, df2, df3), (df1, None, df1), (None, df2, df2)]
)
def test_join_df(set1, set2, expected):
    pd.testing.assert_frame_equal(join_df(set1, set2), expected)


def test_join_df_failure():
    with pytest.raises(AttributeError):
        join_df(df1, "bad value")


@pytest.mark.parametrize(
    "set1, set2, expected", [(df1, None, df1), (None, df2, df2), (df1, df1, df4)]
)
def test_append_df(set1, set2, expected):
    pd.testing.assert_frame_equal(append_df(set1, set2), expected)
